import itertools
import pathlib
import re

import typer

_THIS_FILE = pathlib.Path(__file__)
fp = open(_THIS_FILE.parent / "data" / "word_list.txt")
_WORDS = fp.readlines()
_WORDS = "".join(_WORDS)
fp.close()

app = typer.Typer()


@app.command()
def anagram(word: str, wilds: int = typer.Argument(default=0)) -> list[str]:
    """Return all English anagrams of the input word."""

    letters = list(word)
    n = len(word)

    perms = set(itertools.permutations(letters + wilds * ["[a-z]"], n + wilds))

    perms = r"\b|\b".join("".join(elem) for elem in perms)
    perms = rf"\b{perms}\b"
    patt_1 = re.compile(perms)

    result = set(patt_1.findall(_WORDS))

    print("\n".join(elem for elem in result))

    return list(result)


if __name__ == "__main__":
    app()
    # main()



import deal
class FocusTimeManager:

    @deal.pure
    def __init__(self, client_class):
        self._client = client_class
        self.access_token = ''

    #   ---------------------------------------------------------------------------------------------------------------
    #   Focus Timer Methods

    @deal.pure
    def start(self):
        """Starts the focus timer"""
        pass



"""
Useful time conversion methods.
"""

import datetime

import deal
import pytz

from ..helpers.constants import DATE_FORMAT


@deal.has("import")
def convert_local_time_to_utc(original_time, time_zone: str):
    """
    Converts the datetime object to UTC time. Utilizes the time_zone string for proper conversion.

    Arguments:
        original_time (datetime): Datetime object
        time_zone: Time zone of `original_time`

    Returns:
        datetime: Datetime object with the converted UTC time - with no timezone information attached.

    ??? info "Import Help"
        ```python
        from ticktick.helpers.time_methods import convert_local_time_to_utc
        ```

    ??? Example
        ```python
        pst = datetime(2020, 12, 11, 23, 59)
        converted = convert_local_time_to_utc(pst, 'US/Pacific')
        ```

        ??? success "Result"
            A datetime object that is the UTC equivalent of the original date.

            ```python
            datetime(2020, 12, 12, 7, 59)
            ```
    """

    utc = pytz.utc
    time_zone = pytz.timezone(time_zone)
    original_time = original_time.strftime(DATE_FORMAT)
    time_object = datetime.datetime.strptime(original_time, DATE_FORMAT)
    time_zone_dt = time_zone.localize(time_object)
    return time_zone_dt.astimezone(utc).replace(tzinfo=None)


@deal.pure
def convert_date_to_tick_tick_format(datetime_obj, tz: str):
    """
    Parses ISO 8601 Format to Tick Tick Date Format

    It first converts the datetime object to UTC time based off the passed time zone, and then
    returns a string with the TickTick required date format.

    !!! info Required Format
        ISO 8601 Format Example: 2020-12-23T01:56:07+00:00

        TickTick Required Format: 2020-12-23T01:56:07+0000 -> Where the last colon is removed for timezone

    Arguments:
        datetime_obj (datetime): Datetime object to be parsed.
        tz: Time zone string.

    Returns:
        str: The TickTick accepted date string.

    ??? info "Import Help"
        ```python
        from ticktick.helpers.time_methods import convert_iso_to_tick_tick_format
        ```

    ??? example
        ```python
        date = datetime(2022, 12, 31, 14, 30, 45)
        converted_date = convert_iso_to_tick_tick_format(date, 'US/Pacific')
        ```

        ??? success "Result"
            The proper format for a date string to be used with TickTick dates.

            ```python
            '2022-12-31T22:30:45+0000'
            ```
    """
    date = convert_local_time_to_utc(datetime_obj, tz)
    date = date.replace(tzinfo=datetime.timezone.utc).isoformat()
    date = date[::-1].replace(":", "", 1)[::-1]
    return date



import datetime

# for debugging the daemon
import logging
import os
import random
import subprocess
import sys
import time
from pathlib import Path

import appdirs  # type: ignore[import]

from .loggers.tick_logger import StreamToLogger, log
from .parser.tick_parser import tick_parser  # type: ignore[import]
from .ticktick.api import TickTickClient

# Register App with ticktick servers at https://developer.ticktick.com/manage
from .ticktick.oauth2 import OAuth2

_DEBUG = True

tick_parser.prog = "py -m rob." + Path(__file__).stem

# Execute the parse_args() method
args = tick_parser.parse_args()

_BASE_PATH = Path(appdirs.user_data_dir())

_CLIENT_ID = os.environ.get("TICKTICK_CLIENT_ID", "")
_CLIENT_SECRET = os.environ.get("TICKTICK_CLIENT_SECRET", "")

if not _CLIENT_ID or not _CLIENT_SECRET:
    print(
        "Missing environment variables 'TICKTICK_CLIENT_ID' and/or 'TICKTICK_CLIENT_SECRET'\nRegister App with ticktick servers at https://developer.ticktick.com/manage"
    )
    exit(1)


_TICKTICK_USERNAME = os.environ.get("TICKTICK_USERNAME", "")
_TICKTICK_PASSWORD = os.environ.get("TICKTICK_PASSWORD", "")

if not _TICKTICK_USERNAME or not _TICKTICK_USERNAME:
    print(
        "Missing environment variables 'TICKTICK_USERNAME' and/or 'TICKTICK_PASSWORD'"
    )
    exit(1)

_TOKEN_FILE = _BASE_PATH / "robolson" / "tick" / "cache" / ".token-oauth"

if not _TOKEN_FILE.exists():
    os.makedirs(_TOKEN_FILE.parent, exist_ok=True)

if args.token:
    os.remove(_TOKEN_FILE)
_TASK_FILE = os.environ.get("TICKTICK_TASK_FILE", "")
if not _TASK_FILE:
    _TASK_FILE = _BASE_PATH / "robolson" / "tick" / "tasks.txt"
    if not _TASK_FILE.exists():
        os.makedirs(_TASK_FILE.parent, exist_ok=True)
        open(f"{_TASK_FILE}", "w").close()


if args.get:
    last_modified = datetime.datetime.fromtimestamp(os.path.getmtime(_TASK_FILE))
    if datetime.datetime.today() - last_modified < datetime.timedelta(minutes=60):
        with open(_TASK_FILE, "r") as fp:
            print(", ".join([elem.rstrip() for elem in fp.readlines()]), end="")

    exit(0)


def get_due(tick_client) -> list[str]:
    today_str = datetime.datetime.today().isoformat("!")
    today_date = today_str
    today_str = today_str.split("!")[0]

    today = datetime.datetime.fromisoformat(today_str)

    due = [
        elem
        for elem in tick_client.state["tasks"]
        if "dueDate" in elem.keys()
        and today + datetime.timedelta(days=1)
        > datetime.datetime.fromisoformat(elem["dueDate"][:-5])
    ]

    return due


def main() -> None:

    global args
    if _DEBUG:
        # For Debugging, these will redirect standard streams to the logger
        stdout_logger = logging.getLogger("STDOUT")
        sl = StreamToLogger(stdout_logger, logging.INFO)
        sys.stdout = sl

        stderr_logger = logging.getLogger("STDERR")
        sl = StreamToLogger(stderr_logger, logging.ERROR)
        sys.stderr = sl

    auth_client = OAuth2(
        client_id=_CLIENT_ID,
        client_secret=_CLIENT_SECRET,
        redirect_uri="http://127.0.0.1:8080",
        cache_path=str(_TOKEN_FILE),
        check_cache=True,
    )

    tick_client = TickTickClient(
        username=_TICKTICK_USERNAME, password=_TICKTICK_PASSWORD, oauth=auth_client
    )

    if args.update:
        due = get_due(tick_client)

        with open(f"{_TASK_FILE}", "w") as fp:
            if due:
                fp.write("\n".join([elem["title"] for elem in due]))  # type: ignore[index]
            else:
                fp.write("")

        exit(0)

    if args.daemon:
        subprocess.Popen(["pyw", "-m", "rob.tick"])
        exit(0)

    while True:
        due = get_due(tick_client)

        # Create Overdue Tasks in "me" list for Pavlok
        for task in due:
            # Do not trigger during sleeping hours
            if not 8 < datetime.datetime.now().hour < 20:
                continue

            due_date = datetime.datetime.fromisoformat(task["dueDate"][:-5])
            due_diff = datetime.datetime.now() - due_date

            if due_diff.days > 0 and 0.99**due_diff.days < random.random():

                if task["projectId"] == "54384b29b84562b41688e91a":
                    # If task in Joes List, add to Joes Pavlok List
                    new_projectId = "8999486ebf8fc984f752dd46"

                elif task["projectId"] == "af22459697ae1243ca23f4d9":
                    # If task in Eryns List, add to Eryns Pavlok List
                    new_projectId = "88ba4b50ab907613d84d0cb6"

                else:
                    # Add to Robs Pavlok List
                    new_projectId = "611479cafba2c1d019f96b45"

                new_due = datetime.datetime.fromordinal(
                    datetime.datetime.today().toordinal()
                )
                new_due = new_due + datetime.timedelta(hours=20)
                new_due = new_due.isoformat("T") + "+0000"
                new_task = tick_client.task.builder(
                    title=f"Overdue({due_diff.days}): {task['title']}",
                    dueDate=new_due,
                    content="Created for Pavlok",
                    priority=5,
                    projectId=new_projectId,
                )
                new_task["dueDate"] = new_due
                tick_client.task.create(new_task)

        with open(f"{_TASK_FILE}", "w") as fp:
            if due:
                # write name of due task to file if it is assigned to inbox
                fp.write("\n".join([elem["title"] for elem in due if elem["projectId"] == "inbox115493726"]))  # type: ignore[index]
            else:
                fp.write("")

        # log.info("syncing with ticktick servers")
        tick_client.sync()

        time.sleep(299)


if __name__ == "__main__":
    main()



import argparse

tick_parser = argparse.ArgumentParser(
    prog="py -m rob.tick",
    allow_abbrev=True,
    add_help=True,
    # usage="$(prog)s [-h] path",
    description="Periodically syncs a local cache with ticktick servers.",
    epilog="(C) Rob",
)

commands = tick_parser.add_mutually_exclusive_group()

front_options = tick_parser.add_argument_group()

commands.add_argument(
    "-d",
    "--daemon",
    action="store_true",
    default=False,
    help="Run this as a background process.",
)

commands.add_argument(
    "-g", "--get", action="store_true", default=False, help="Get today's tasks."
)


front_options.add_argument(
    "-t",
    "--token",
    action="store_true",
    default=False,
    help="Force an OAuth Token update.",
)


commands.add_argument(
    "-u", "--update", action="store_true", default=False, help="Update the task cache."
)



import datetime
from calendar import monthrange

import pytz

from ..helpers.constants import DATE_FORMAT
from ..helpers.time_methods import (
    convert_date_to_tick_tick_format,
    convert_local_time_to_utc,
)


class TaskManager:
    """
    Handles all interactions for tasks.
    """

    TASK_CREATE_ENDPOINT = "/open/v1/task"

    def __init__(self, client_class):
        # ._client is a reference to the original client_class
        self._client = client_class

        # declare access token
        self.oauth_access_token = ""

        # set access token to valid oauth access token if available
        if self._client.oauth_manager.access_token_info is not None:
            self.oauth_access_token = self._client.oauth_manager.access_token_info[
                "access_token"
            ]

        # oauth headers have some extra fields
        self.oauth_headers = {
            "Content-Type": "application/json",
            "Authorization": "Bearer {}".format(self.oauth_access_token),
            "User-Agent": self._client.USER_AGENT,
        }

        self.headers = self._client.HEADERS

    def _generate_create_url(self):
        """
        Returns the endpoint url for task creation
        """

        CREATE_ENDPOINT = "/open/v1/task"
        return self._client.OPEN_API_BASE_URL + CREATE_ENDPOINT

    def create(self, task):
        """
        Create a task. Use [`builder`][managers.tasks.TaskManager.builder] for easy task dictionary
        creation.

        !!! warning
            Creating tasks with tags is not functional but will be implemented in a future update.

        Arguments:
            task (dict): Task dictionary to be created.

        Returns:
            dict: Dictionary of created task object. Note that the task object is a "simplified" version of the full
            task object. Use [`get_by_id`][api.TickTickClient.get_by_id] for the full task object.

        !!! example "Creating Tasks"
            === "Just A Name"
                ```python
                title = "Molly's Birthday"
                task = client.task.builder(title)   # Local dictionary
                molly = client.task.create(task)    # Create task remotely
                ```

                ??? success "Result"

                    ```python
                    {'id': '60ca9dbc8f08516d9dd56324',
                     'projectId': 'inbox115781412',
                     'title': "Molly's Birthday",
                     'timeZone': '',
                     'reminders': [],
                     'priority': 0,
                     'status': 0,
                     'sortOrder': -1336456383561728,
                     'items': []}
                    ```
                    ![image](https://user-images.githubusercontent.com/56806733/122314079-5898ef00-cecc-11eb-8614-72b070b306c6.png)

            === "Dates and Descriptions"
                ```python
                title = "Molly's Birthday Party"

                start_time = datetime(2027, 7, 5, 14, 30)  # 7/5/2027 @ 2:30PM
                end_time = datetime(2027, 7, 5, 19, 30)    # 7/5/2027 @ 7:30PM

                content = "Bring Cake"

                task = client.task.builder(title,
                                           startDate=start_time,
                                           dueDate=end_time,
                                           content=content)

                mollys_party = client.task.create(task)
                ```

                ??? success "Result"
                    ```python
                    {'id': '60ca9fe58f08fe31011862f2',
                     'projectId': 'inbox115781412',
                     'title': "Molly's Birthday Party",
                     'content': 'Bring Cake',
                     'timeZone': '',
                     'startDate': '2027-07-05T21:30:00.000+0000',
                     'dueDate': '2027-07-06T02:30:00.000+0000',
                     'priority': 0,
                     'status': 0,
                     'sortOrder': -1337555895189504,
                     'items': [],
                     'allDay': False}
                    ```
                    ![image](https://user-images.githubusercontent.com/56806733/122314760-a4986380-cecd-11eb-88af-9562d352470f.png)

            === "Different Project"
                ```python
                # Get the project object
                events = client.get_by_fields(name="Events", search='projects')

                events_id = events['id']    # Need the project object id

                title = "Molly's Birthday"

                task = client.task.builder(title, projectId=events_id)

                mollys_birthday = client.task.create(task)
                ```

                ??? success "Result"
                    ```python
                    {'id': '60caa2278f08fe3101187002',
                     'projectId': '60caa20d8f08fe3101186f74',
                     'title': "Molly's Birthday",
                     'timeZone': '',
                     'reminders': [],
                     'priority': 0,
                     'status': 0,
                     'sortOrder': -1099511627776,
                     'items': []}
                    ```
                    ![image](https://user-images.githubusercontent.com/56806733/122315454-eece1480-cece-11eb-8394-94a2aec1ba70.png)
        """

        # generate url
        url = self._generate_create_url()

        # make request
        response = self._client.http_post(
            url=url, json=task, headers=self.oauth_headers
        )

        # sync local state
        self._client.sync()

        # TODO: Figure out tags
        # since the openapi does not explicitly support tag creation - lets create a new tag for the new task
        # try:
        #     tags = task['tag']
        #     if isinstance(tags, str):
        #         # single tag -> check to see if it exists before creating another one
        #         search = self._client.get_by_fields(name=tags, search='tags')
        #     elif:
        #         isinstance(tags, list)
        #         # multiple tags
        #         pass
        # except KeyError:
        #     pass

        # set 'inbox' to be the actual inbox id
        if response["projectId"] == "inbox":
            response["projectId"] = self._client.inbox_id

        # return response
        return response

    def _generate_update_url(self, taskID: str):
        """
        Generates the url for updating a task based off the taskID
        """

        UPDATE_ENDPOINT = f"/open/v1/task/{taskID}"
        return self._client.OPEN_API_BASE_URL + UPDATE_ENDPOINT

    def update(self, task):
        """
        Update a task. The task should already be created.

        To update a task, change any field in it's dictionary directly then pass to the method.

        !!! warning
            Creating tasks with tags is not functional but will be implemented in a future update.

        Arguments:
            task (dict): Task dictionary to be updated

        Returns:
            dict: The updated task dictionary object

        !!! tip "Formatting Dates Help"
            TickTick uses a certain syntax for their dates. To convert a datetime object to a compatible
            string to be used for updating dates, see [convert_date_to_tick_tick_format][helpers.time_methods.convert_date_to_tick_tick_format]

        !!! example "Changing The Date"

            ```python
            # Get the task
            mollys_birthday = client.get_by_fields(title="Molly's Birthday", search="tasks")

            # New Date
            new_date = datetime(2027, 5, 6)

            # Get the new date string
            new_date_string = convert_date_to_tick_tick_format(new_date, tz=client.time_zone)

            # Update the task dictionary
            mollys_birthday['startDate'] = new_date_string

            # Update the task
            molly_updated = client.task.update(mollys_birthday)
            ```

            **Original Task**
            ![image](https://user-images.githubusercontent.com/56806733/122316205-49b43b80-ced0-11eb-8d14-61fa5bb5b10a.png)

            ??? success "Result"
                ```python
                {'id': '60caa2278f08fe3101187002',
                'projectId': '60caa20d8f08fe3101186f74',
                'title': "Molly's Birthday",
                'content': '',
                'timeZone': '',
                'startDate': '2027-05-06T07:00:00.000+0000',
                'dueDate': '2027-05-05T07:00:00.000+0000',
                'reminders': [],
                'priority': 0,
                'status': 0,
                'sortOrder': -1099511627776,
                'items': [],
                'allDay': True}
                ```
                ![image](https://user-images.githubusercontent.com/56806733/122316408-ad3e6900-ced0-11eb-89f9-6d980b5cc954.png)
        """

        # TODO: Make tags work

        # generate url
        url = self._generate_update_url(task["id"])

        # make request
        response = self._client.http_post(
            url=url, json=task, headers=self.oauth_headers
        )

        # sync local state
        self._client.sync()

        # return response
        return response

    def _generate_mark_complete_url(self, projectID, taskID):
        """
        Generates the url for marking a task as complete based off the projectID and taskID
        """

        COMPLETE_ENDPOINT = f"/open/v1/project/{projectID}/task/{taskID}/complete"
        return self._client.OPEN_API_BASE_URL + COMPLETE_ENDPOINT

    def complete(self, task: dict):
        """
        Marks a task as complete. Pass in the task dictionary to be marked as completed.

        !!! note
            The task should already be created

        Arguments:
            task (dict): The task dictionary object.

        Returns:
            dict: The original passed in task.

        !!! example "Task Completing"
            ```python
            # Lets assume that we have a task named "Dentist" that we want to mark as complete.

            dentist_task = client.get_by_fields(title='Dentist', search='tasks')
            complete_task = client.task.complete(dentist_task)  # Pass the task dictionary
            ```

            ??? success "Result"
                The task is completed and the dictionary object returned.

                ```python
                {'id': '5fff5009b04b355792c79397', 'projectId': 'inbox115781412', 'sortOrder': -99230924406784,
                'title': 'Go To Dentist', 'content': '', 'startDate': '2021-01-13T08:00:00.000+0000',
                'dueDate': '2021-01-13T08:00:00.000+0000', 'timeZone': 'America/Los_Angeles', 'isFloating': False,
                'isAllDay': True, 'reminders': [], 'exDate': [], 'priority': 0, 'status': 2, 'items': [],
                'progress': 0, 'modifiedTime': '2021-01-13T19:56:11.000+0000', 'etag': 'djiiqso6', 'deleted': 0,
                'createdTime': '2021-01-13T19:54:49.000+0000', 'creator': 6147345572, 'kind': 'TEXT'}
                ```

                **Before**

                ![image](https://user-images.githubusercontent.com/56806733/104503673-39510b00-5596-11eb-88df-88eeee9ab4b0.png)

                **After**

                ![image](https://user-images.githubusercontent.com/56806733/104504069-c4ca9c00-5596-11eb-96c9-5698e19989ea.png)
        """

        # generate url
        url = self._generate_mark_complete_url(task["projectId"], task["id"])

        # make request
        response = self._client.http_post(
            url=url, json=task, headers=self.oauth_headers
        )

        # sync local state
        self._client.sync()

        if response == "":
            return task

        # return response
        return response

    def _generate_delete_url(self):
        """
        Generates end point url for task deletion
        """
        return self._client.BASE_URL + "batch/task"

    def delete(self, task):
        """
        Deletes a task. Supports single task deletion, and batch task deletion.

        For a single task pass in the task dictionary. For multiple tasks pass in a list of task dictionaries.

        Arguments:
             task (str or list):
                 **Single Task (dict)**: Task dictionary to be deleted

                 **Multiple Tasks (list)**: List of task dictionaries to be deleted

        Returns:
             dict or list:
                **Single Task (dict)**: Task dictionary that was deleted

                **Multiple Tasks (list)**: List of task dictionaries that were deleted

        !!! example "Task Deletion"

            === "Single Task Deletion"

                ```python
                # Get the task
                task = client.get_by_fields(title="Molly's Birthday", search="tasks")

                # Delete the task
                deleted = client.task.delete(task)
                ```

                ??? success "Result"
                    ``` python
                    {'id': '60caa2278f08fe3101187002',
                    'projectId': '60caa20d8f08fe3101186f74',
                    'sortOrder': -1099511627776,
                    'title': "Molly's Birthday",
                    'content': '',
                    'startDate': '2027-05-06T07:00:00.000+0000',
                    'dueDate': '2027-05-06T07:00:00.000+0000',
                    'timeZone': '',
                    'isFloating': False,
                    'isAllDay': True,
                    'reminders': [],
                    'repeatFirstDate': '2027-05-05T07:00:00.000+0000',
                    'exDate': [],
                    'priority': 0,
                    'status': 0,
                    'items': [],
                    'progress': 0,
                    'modifiedTime': '2021-06-17T01:25:19.000+0000',
                    'etag': 'rrn4paqp',
                    'deleted': 0,
                    'createdTime': '2021-06-17T01:15:19.365+0000',
                    'creator': 119784412,
                    'kind': 'TEXT'}
                    ```

            === "Multiple Task Deletion"
                ``` python
                # Get the tasks
                wash_car = client.get_by_fields(title="Wash Car", search="tasks")
                do_dishes = client.get_by_fields(title="Do Dishes", search="tasks")

                # Make a list for the tasks
                to_delete = [wash_car, do_dishes]

                # Delete the tasks
                deleted = client.task.delete(to_delete)
                ```

                **Before**

                ![image](https://user-images.githubusercontent.com/56806733/122317746-e11a8e00-ced2-11eb-8449-519615de5935.png)


                ??? success "Result"
                    ```python
                    [{'id': '60caa8e714f7103cef35765a', 'projectId': '60caa20d8f08fe3101186f74',
                    'sortOrder': -1099511627776, 'title': 'Wash Car', 'content': '',
                    'timeZone': 'America/Los_Angeles', 'isFloating': False,
                    'reminder': '', 'reminders': [], 'exDate': [], 'priority': 0, 'status': 0,
                    'items': [], 'progress': 0, 'modifiedTime': '2021-06-17T01:44:07.000+0000', 'etag': '8372m61k',
                    'deleted': 0, 'createdTime': '2021-06-17T01:44:07.000+0000',
                    'creator': 115761422, 'tags': [], 'kind': 'TEXT'},

                    {'id': '60caa8ea14f7103cef35765f', 'projectId': '60caa20d8f08fe3101186f74',
                    'sortOrder': -2199023255552, 'title': 'Do Dishes', 'content': '',
                    'timeZone': 'America/Los_Angeles', 'isFloating': False, 'reminder': '',
                    'reminders': [], 'exDate': [], 'priority': 0, 'status': 0, 'items': [],
                    'progress': 0, 'modifiedTime': '2021-06-17T01:44:10.000+0000',
                    'etag': 'sfka0mvn', 'deleted': 0, 'createdTime': '2021-06-17T01:44:10.000+0000',
                    'creator': 1155481312, 'tags': [], 'kind': 'TEXT'}]
                    ```
                    ![image](https://user-images.githubusercontent.com/56806733/122317923-2212a280-ced3-11eb-8a6b-8a32fa8426ce.png)


        """

        # generate url
        url = self._generate_delete_url()

        to_delete = []

        # if its just a dict then we are going to have to make a list object for it
        if isinstance(task, dict):
            # ticktick returns for the 'projectId': 'inbox' instead of the actual inbox id - which is required for
            # proper deletion
            if task["projectId"] == "inbox":
                task["projectId"] = self._client.inbox_id

            delete_dict = {"projectId": task["projectId"], "taskId": task["id"]}
            to_delete.append(delete_dict)

        # iterate through "task"
        else:
            for item in task:
                if item["projectId"] == "inbox":
                    item["projectId"] = self._client.inbox_id
                delete_dict = {"projectId": item["projectId"], "taskId": item["id"]}
                to_delete.append(delete_dict)

        payload = {"delete": to_delete}
        # make request
        self._client.http_post(
            url, json=payload, cookies=self._client.cookies, headers=self.headers
        )

        # sync local state
        self._client.sync()

        # return input
        return task

    def make_subtask(self, obj, parent: str):
        """
        Makes the passed task(s) sub-tasks to the parent task.

        !!! note "Important"
            All of the tasks should already be created prior to using this method. Furthermore,
            the tasks should already be present in the same project as the parent task.

        Arguments:
            obj (dict):
                **Single Sub-Task (dict)**: The task object dictionary.

                **Multiple Sub-Tasks (list)**: A list of task object dictionaries.

            parent (str): The ID of the task that will be the parent task.

        Returns:
            dict:
             **Single Sub-Task (dict)**: Created sub-task dictionary.

             **Multiple Sub-Tasks (list)**: List of created sub-task dictionaries.

        Raises:
            TypeError: `obj` must be a dictionary or list of dictionaries. `parent` must be a string.
            ValueError: If `parent` task doesn't exist.
            ValueError: If `obj` does not share the same project as parent.
            RuntimeError: If the creation was unsuccessful.

        !!! example "Creating Sub-Tasks"
            === "Single Sub-Task Creation"
                Pass the task object that will be made a sub-task to the parent with the passed ID.

                ```python
                # Lets make a task in our inbox named "Read" with a sub-task "50 Pages"
                read_task = client.task.create('Read')
                pages_task = client.task.create('50 pages')
                now_subtask = client.task.make_subtask(pages_task, read_task['id'])
                ```

                ??? success "Result"
                    The dictionary of the sub-task is returned.

                    ```python
                    {'id': '5ffff4968f08af50b4654c6b', 'projectId': 'inbox115781412', 'sortOrder': -3298534883328,
                    'title': '50 pages', 'content': '', 'timeZone': 'America/Los_Angeles', 'isFloating': False,
                    'reminder': '', 'reminders': [], 'priority': 0, 'status': 0, 'items': [],
                    'modifiedTime': '2021-01-14T07:37:36.487+0000', 'etag': 'xv5cjzoz', 'deleted': 0,
                    'createdTime': '2021-01-14T07:36:54.751+0000', 'creator': 115781412,
                    'parentId': '5ffff4968f08af50b4654c62', 'kind': 'TEXT'}
                    ```

                    **Before**

                    ![image](https://user-images.githubusercontent.com/56806733/104558809-4272c400-55f8-11eb-8c55-e2f77c9d1ac8.png)

                    **After**

                    ![image](https://user-images.githubusercontent.com/56806733/104558849-55859400-55f8-11eb-9692-c3e01aa73233.png)

            === "Multiple Sub-Task Creation"
                Pass all the tasks you want to make sub-tasks in a list.

                ```python
                # Lets make a task in our inbox named "Read" with a sub-tasks "50 Pages", "100 Pages", and "200 Pages"
                read_task = client.task.create("Read")
                # Lets batch create our sub-tasks
                fifty_pages = client.task.builder('50 Pages')
                hundred_pages = client.task.builder('100 Pages')
                two_hundred_pages = client.task.builder('200 Pages')
                page_tasks = client.task.create([fifty_pages, hundred_pages, two_hundred_pages])
                # Make the page tasks sub-tasks to read_task
                subtasks = client.task.make_subtask(page_tasks, read_task['id'])
                ```

                ??? success "Result"
                    A list of the sub-tasks is returned.

                    ```python
                    [{'id': '5ffff6348f082c11cc0da84d', 'projectId': 'inbox115781412', 'sortOrder': -5497558138880,
                    'title': '50 Pages', 'content': '', 'timeZone': 'America/Los_Angeles',
                    'isFloating': False, 'reminder': '', 'reminders': [], 'priority': 0, 'status': 0,
                    'items': [], 'modifiedTime': '2021-01-14T07:45:04.032+0000', 'etag': 'avqm3u6o',
                    'deleted': 0, 'createdTime': '2021-01-14T07:43:48.858+0000', 'creator': 567893575,
                    'parentId': '5ffff6348f082c11cc0da84a', 'kind': 'TEXT'},

                    {'id': '5ffff6348f082c11cc0da84e', 'projectId': 'inbox115781412', 'sortOrder': -5497558138880,
                    'title': '100 Pages', 'content': '', 'timeZone': 'America/Los_Angeles',
                    'isFloating': False, 'reminder': '', 'reminders': [], 'priority': 0, 'status': 0,
                    'items': [], 'modifiedTime': '2021-01-14T07:45:04.035+0000', 'etag': '6295mmmu',
                    'deleted': 0, 'createdTime': '2021-01-14T07:43:49.286+0000', 'creator': 567893575,
                    'parentId': '5ffff6348f082c11cc0da84a', 'kind': 'TEXT'},

                    {'id': '5ffff6348f082c11cc0da84f', 'projectId': 'inbox115781412', 'sortOrder': -5497558138880,
                    'title': '200 Pages', 'content': '', 'timeZone': 'America/Los_Angeles',
                    'isFloating': False, 'reminder': '', 'reminders': [], 'priority': 0, 'status': 0,
                    'items': [], 'modifiedTime': '2021-01-14T07:45:04.038+0000', 'etag': 'du59zwck',
                    'deleted': 0, 'createdTime': '2021-01-14T07:43:49.315+0000', 'creator': 567893575,
                    'parentId': '5ffff6348f082c11cc0da84a', 'kind': 'TEXT'}]
                    ```

                    **Before**

                    ![image](https://user-images.githubusercontent.com/56806733/104559418-36d3cd00-55f9-11eb-9004-177671a92474.png)

                    **After**

                    ![image](https://user-images.githubusercontent.com/56806733/104559535-64207b00-55f9-11eb-84cf-ca4f989ea075.png)
        """
        if not isinstance(obj, dict) and not isinstance(obj, list):
            raise TypeError("obj must be a dictionary or list of dictionaries")

        if not isinstance(parent, str):
            raise TypeError("parent must be a string")

        if isinstance(obj, dict):
            obj = [obj]

        parent_obj = self._client.get_by_id(search="tasks", obj_id=parent)
        if not parent_obj:
            raise ValueError("Parent task must exist before creating sub-tasks")

        ids = []
        # Go through obj and if the projects are different make them the same as parent
        for o in obj:
            if o["projectId"] != parent_obj["projectId"]:
                raise ValueError("All tasks must be in the same project as the parent")
            ids.append(o["id"])

        subtasks = []
        for i in ids:  # Create the object dictionaries for setting the subtask
            temp = {
                "parentId": parent,
                "projectId": parent_obj["projectId"],
                "taskId": i,
            }
            subtasks.append(temp)

        url = self._client.BASE_URL + "batch/taskParent"
        response = self._client.http_post(
            url, json=subtasks, cookies=self._client.cookies, headers=self.headers
        )
        self._client.sync()
        # Find and return the updated child objects
        subtasks = []
        for task_id in ids:
            subtasks.append(self._client.get_by_id(task_id, search="tasks"))
        if len(subtasks) == 1:
            return subtasks[0]  # Return just the dictionary object if its a single task
        else:
            return subtasks

    def move(self, obj, new: str):
        """
        Moves task(s) from their current project to the new project. It will move the specified
        tasks with `obj` to the new project.

        !!! important
            If moving multiple tasks, they must all be from the same project.
        Arguments:
            obj (dict or list):
                **Single Task (dict)**: Pass the single task dictionary object to move.

                **Multiple Tasks (list)**: Pass a list of task dictionary objects to move.
            new: The ID string of the project that the task(s) should be moved to.

        Returns:
            dict or list:
            **Single Task (dict)**: Returns the dictionary of the moved task.

            **Multiple Tasks (list)**: Returns a list of dictionaries for the moved tasks.

        Raises:
            TypeError: If `obj` is not a dict or list or if `new` is not a str.
            ValueError: For multiple tasks, if the projects are not all the same.
            ValueError: If the new project does not exist.
            RuntimeError: If the task(s) could not be successfully moved.

        !!! example "Move Examples"
            === "Moving A Single Task"
                Pass in the task object, and the ID of the project the task should be moved to.

                ```python
                # Lets assume that we have a task 'Read' that exists in a project named "Work"
                # Lets move that task to the inbox
                read_task = client.get_by_fields(title='Read', search='tasks')
                move_read_task = client.task.move(read_task, client.inbox_id)
                ```

                ??? success "Result"
                    The dictionary object of the moved task is returned.

                    ```python
                    {'id': '5fffed61b04b355792c799a8', 'projectId': 'inbox115781412', 'sortOrder': 0,
                    'title': 'Read', 'content': '', 'startDate': '2021-01-13T08:00:00.000+0000',
                    'dueDate': '2021-01-13T08:00:00.000+0000', 'timeZone': 'America/Los_Angeles',
                    'isFloating': False, 'isAllDay': True, 'reminders': [], 'exDate': [], 'priority': 0,
                    'status': 0, 'items': [], 'progress': 0, 'modifiedTime': '2021-01-14T07:08:15.875+0000',
                    'etag': 'twrmcr55', 'deleted': 0, 'createdTime': '2021-01-14T07:06:09.000+0000',
                    'creator': 47593756, 'tags': [], 'kind': 'TEXT'}
                    ```

                    **Before**

                    ![image](https://user-images.githubusercontent.com/56806733/104556170-f1f96780-55f3-11eb-9a35-aecc3beea105.png)

                    **After**

                    ![image](https://user-images.githubusercontent.com/56806733/104556336-46044c00-55f4-11eb-98c1-4cffcf4bd006.png)

            === "Moving Multiple Tasks"
                Pass in the task objects in a list, and the ID of the project that tasks should be moved to.
                Again, the tasks in the list should all be from the same project.

                ```python
                # Lets move two tasks: 'Read' and 'Write' that exist in a project named "Work"
                # Lets move the tasks to another project named "Hobbies" that already exists.
                hobbies_project = client.get_by_fields(name='Hobbies', search='projects')
                hobbies_id = hobbies_project['id']  # Id of the hobbies project
                read_task = client.get_by_fields(title='Read', search='tasks')
                write_task = client.get_by_fields(title='Write', search='tasks')
                move_tasks = client.task.move([read_task, write_task], hobbies_id)  # Task objects in a list
                ```

                ??? success "Result"
                    The tasks that were moved are returned in a list.

                    ```python
                    [{'id': '5ffff003b04b355792c799d3', 'projectId': '5fffeff68f08654c982c141a', 'sortOrder': 0,
                    'title': 'Read', 'content': '', 'startDate': '2021-01-13T08:00:00.000+0000',
                    'dueDate': '2021-01-13T08:00:00.000+0000', 'timeZone': 'America/Los_Angeles',
                    'isFloating': False, 'isAllDay': True, 'reminders': [], 'exDate': [], 'priority': 0,
                    'status': 0, 'items': [], 'progress': 0, 'modifiedTime': '2021-01-14T07:19:28.595+0000',
                    'etag': 'co8jfqyn', 'deleted': 0, 'createdTime': '2021-01-14T07:17:23.000+0000',
                    'creator': 768495743, 'kind': 'TEXT'},

                    {'id': '5ffff004b04b355792c799d4', 'projectId': '5fffeff68f08654c982c141a', 'sortOrder': 0,
                    'title': 'Write', 'content': '', 'startDate': '2021-01-13T08:00:00.000+0000',
                    'dueDate': '2021-01-13T08:00:00.000+0000', 'timeZone': 'America/Los_Angeles',
                    'isFloating': False, 'isAllDay': True, 'reminders': [], 'exDate': [], 'priority': 0,
                    'status': 0, 'items': [], 'progress': 0, 'modifiedTime': '2021-01-14T07:19:28.596+0000',
                    'etag': '5unkf7xg', 'deleted': 0, 'createdTime': '2021-01-14T07:17:24.000+0000',
                    'creator': 768495743, 'tags': [], 'kind': 'TEXT'}]
                    ```

                    **Before**

                    ![image](https://user-images.githubusercontent.com/56806733/104557103-857f6800-55f5-11eb-8b92-cf51bc159745.png)

                    **After**

                    ![image](https://user-images.githubusercontent.com/56806733/104557388-063e6400-55f6-11eb-8ba4-aa64f3f739bd.png)

        """
        # Type errors
        if not isinstance(obj, dict) and not isinstance(obj, list):
            raise TypeError("obj should be a dict or list of dicts")
        if not isinstance(new, str):
            raise TypeError("new should be a string")

        # Get the parent project
        if new != self._client.inbox_id:
            project = self._client.get_by_id(new, search="projects")
            if not project:
                raise ValueError("The ID for the new project does not exist")

        if isinstance(obj, dict):
            obj = [obj]

        # Go through and check that the projects are all the same
        move_tasks = []
        project_id = obj[0]["projectId"]
        for task in obj:
            if task["projectId"] != project_id:
                raise ValueError("All the tasks must come from the same project")
            else:
                move_tasks.append(
                    {
                        "fromProjectId": project_id,
                        "taskId": task["id"],
                        "toProjectId": new,
                    }
                )

        url = self._client.BASE_URL + "batch/taskProject"
        self._client.http_post(
            url, json=move_tasks, cookies=self._client.cookies, headers=self.headers
        )
        self._client.sync()
        # Return the tasks in the new list
        ids = [x["id"] for x in obj]
        return_list = []
        for i in ids:
            return_list.append(self._client.get_by_id(i))
        if len(return_list) == 1:
            return return_list[0]
        else:
            return return_list

    def move_all(self, old: str, new: str) -> list:
        """
        Moves all the tasks from the old project to the new project.

        Arguments:
            old: ID of the old project.
            new: ID of the new project.

        Returns:
            The tasks contained in the new project.

        Raises:
            ValueError: If either the old or new projects do not exist.
            RuntimeError: If the movement was unsuccessful.

        !!! example
            Lets assume that we have a project named "School", and another project named "Work". To move all the tasks from "School" to "Work":

            ```python
            # Get the projects
            school_project = client.get_by_fields(name='School', search='projects')
            work_project = client.get_by_fields(name='Work', search='projects')
            # Call the method
            moved_tasks = client.task.move_all(school_project['id'], work_project['id'])
            ```

            ??? success "Result"
                The tasks that were moved are returned.

                ```python
                [{'id': '5ffea9afe4b062d60dd62aef', 'projectId': '5ffea9afe4b062d60dd62aea', 'sortOrder': 0,
                'title': 'Finish documentation for project', 'content': '', 'timeZone': 'America/Los_Angeles',
                'isFloating': False, 'reminder': '', 'reminders': [], 'priority': 0, 'status': 0, 'items': [],
                'modifiedTime': '2021-01-13T08:06:31.407+0000', 'etag': 'ogclghmd', 'deleted': 0,
                'createdTime': '2021-01-13T08:05:03.901+0000', 'creator': 447666584, 'kind': 'TEXT'},

                {'id': '5ffea9b0e4b062d60dd62af4', 'projectId': '5ffea9afe4b062d60dd62aea', 'sortOrder': 0,
                'title': 'Call the boss man', 'content': '', 'timeZone': 'America/Los_Angeles',
                'isFloating': False, 'reminder': '', 'reminders': [], 'priority': 0, 'status': 0, 'items': [],
                'modifiedTime': '2021-01-13T08:06:31.409+0000', 'etag': '65c73q8i', 'deleted': 0,
                'createdTime': '2021-01-13T08:05:04.117+0000', 'creator': 447666584, 'kind': 'TEXT'}]
                ```

                **Before**: Two tasks are contained in the "School" project

                ![image](https://user-images.githubusercontent.com/56806733/104423574-1e997a80-5533-11eb-9417-34c31e603d21.png)

                **After**: The two tasks are moved to the 'Work' project

                ![image](https://user-images.githubusercontent.com/56806733/104423710-4a1c6500-5533-11eb-90f3-2c3d024280af.png)
        """
        # Make sure that old and new id's exist
        if old != self._client.inbox_id:
            old_list = self._client.get_by_fields(id=old, search="projects")
            if not old_list:
                raise ValueError(f"Project Id '{old}' Does Not Exist")

        if new != self._client.inbox_id:
            new_list = self._client.get_by_fields(id=new, search="projects")
            if not new_list:
                raise ValueError(f"Project Id '{new}' Does Not Exist")

        # Get the tasks from the old list
        tasks = self.get_from_project(old)
        if not tasks:
            return tasks  # No tasks to move so just return the empty list
        task_project = []  # List containing all the tasks that will be updated

        for task in tasks:
            task_project.append(
                {"fromProjectId": old, "taskId": task["id"], "toProjectId": new}
            )

        url = self._client.BASE_URL + "batch/taskProject"
        # Make the initial call to move the tasks
        self._client.http_post(
            url, json=task_project, cookies=self._client.cookies, headers=self.headers
        )

        self._client.sync()
        # Return the tasks in the new list
        return self._client.task.get_from_project(new)

    def get_from_project(self, project: str):
        """
        Obtains the tasks that are contained in the project.

        Arguments:
            project: ID string of the project to get the tasks from.

        Returns:
            dict or list:
            **Single Task In Project (dict)**: The single task object dictionary.

            **Multiple Tasks In Project (list)**: A list of task object dictionaries.

            **No Tasks Found (list)**: Empty list.

        Raises:
            ValueError: If the project ID does not exist.

        !!! example "Getting Uncompleted Tasks From The Inbox"
            ```python
            tasks = client.task.get_from_project(client.inbox_id)
            ```

            ??? success "Result"
                See `Returns` for the different return values based on the amount of tasks present
                in the project.

                ```python
                [{'id': '5ffe93efb04b35082bbce7af', 'projectId': 'inbox115781412', 'sortOrder': 2199023255552, 'title': 'Go To Library',
                'content': '', 'startDate': '2021-01-12T08:00:00.000+0000', 'dueDate': '2021-01-12T08:00:00.000+0000',
                'timeZone': 'America/Los_Angeles', 'isFloating': False, 'isAllDay': True,
                'reminders': [], 'exDate': [], 'priority': 0, 'status': 0, 'items': [], 'progress': 0,
                'modifiedTime': '2021-01-13T06:32:15.000+0000', 'etag': 'kkh0w1jk', 'deleted': 0,
                'createdTime': '2021-01-13T06:32:15.000+0000', 'creator': 447666584, 'tags': [],
                'kind': 'TEXT'},

                {'id': '5ffe93f3b04b35082bbce7b0', 'projectId': 'inbox115781412', 'sortOrder': 1099511627776, 'title': 'Deposit Funds',
                'content': '', 'startDate': '2021-01-12T08:00:00.000+0000', 'dueDate': '2021-01-12T08:00:00.000+0000',
                'timeZone': 'America/Los_Angeles', 'isFloating': False, 'isAllDay': True,
                'reminders': [], 'exDate': [], 'priority': 0, 'status': 0, 'items': [], 'progress': 0, 'modifiedTime': '2021-01-13T06:32:19.000+0000',
                'etag': 'w4hj21wf', 'deleted': 0, 'createdTime': '2021-01-13T06:32:19.000+0000', 'creator': 447666584, 'tags': [],
                'kind': 'TEXT'}]
                ```

                ![image](https://user-images.githubusercontent.com/56806733/104415494-f86ddd80-5526-11eb-8b84-75bf3886ba46.png)
        """
        # Make sure the project exists
        if project != self._client.inbox_id:
            obj = self._client.get_by_fields(id=project, search="projects")
            if not obj:
                raise ValueError(f"List Id '{project}' Does Not Exist")

        # Get the list of tasks that share the project id
        tasks = self._client.get_by_fields(projectId=project, search="tasks")
        if isinstance(tasks, dict):
            return [tasks]
        else:
            return tasks

    def get_completed(self, start, end=None, full: bool = True, tz: str = None) -> list:
        """
        Obtains all completed tasks from the given start date and end date.

        !!! note
            There is a limit of 100 items for the request

        Arguments:
            start (datetime): Start time datetime object.
            end (datetime): End time datetime object.
            full: Boolean specifying whether hours, minutes, and seconds are to be taken into account for the query.
            tz: String specifying a specific time zone, however this will default to your accounts normal time zone.

        Returns:
            A list containing all the completed tasks based on the times.

        Raises:
            TypeError: If the proper types are not used.
            ValueError: If start occurs after end.
            KeyError: If the time zone string passed is not a valid time zone string.
            RuntimeError: If getting the tasks is unsuccessful.

        !!! example "Getting Completed Tasks"
            === "Completed Tasks In A Single Day"
                Getting the tasks for a full, complete day requires passing in
                the datetime object corresponding to the day that you want.

                ```python
                # Get the tasks for 1/11/2021
                tasks = client.task.get_completed(datetime(2021, 1, 11))
                ```

                ??? success "Result"
                    The list of completed tasks is returned.

                    ```python
                    [{'id': '5ffca35f4c201114702a0607', 'projectId': '004847faa60015487be444cb',
                    'sortOrder': -50027779063826, 'title': 'Shoulders and Arms', 'content': '', 'desc': '',
                    'startDate': '2021-01-11T08:00:00.000+0000', 'dueDate': '2021-01-11T08:00:00.000+0000',
                    'timeZone': 'America/Los_Angeles', 'isFloating': False, 'isAllDay': True, 'reminders': [],
                    'repeatFlag': '', 'exDate': [], 'completedTime': '2021-01-11T23:25:46.000+0000',
                    'completedUserId': 185769383, 'priority': 0, 'status': 2, 'items': [], 'progress': 0,
                    'modifiedTime': '2021-01-11T23:25:41.000+0000', 'etag': '6hlk4e8t', 'deleted': 0,
                    'createdTime': '2021-01-11T19:13:35.000+0000', 'creator': 185769383, 'tags': ['fitness'],
                    'commentCount': 0, 'pomodoroSummaries': [{'userId': 185769383, 'count': 0, 'estimatedPomo': 0,
                    'duration': 0}], 'focusSummaries': [{'userId': 185769383, 'pomoCount': 0, 'estimatedPomo': 0,
                    'estimatedDuration': 0, 'pomoDuration': 0, 'stopwatchDuration': 3720}], 'kind': 'TEXT'}]
                    ```

                    ![image](https://user-images.githubusercontent.com/56806733/104562952-e1e68580-55fd-11eb-9e09-f432caa8616b.png)

            === "Completed Tasks Over A Range Of Days"
                Getting the tasks for a range of days requires passing in datetime objects
                for the start day, and the end day that you want.

                ```python
                # Get the tasks between 8/7/18 and 8/10/18
                start = datetime(2018, 8, 7)
                end = datetime(2018, 8, 10)
                tasks = client.task.get_completed(start, end)
                ```

                ??? success "Result"
                    Completed tasks in a list are returned.

                    ```python
                    [{'id': '5ffffebab04b355792c79e38', 'projectId': 'inbox115781412', 'sortOrder': -7696581394432,
                    'title': 'Ride Bike', 'content': '', 'startDate': '2021-01-14T08:00:00.000+0000',
                    'dueDate': '2021-01-14T08:00:00.000+0000', 'timeZone': 'America/Los_Angeles',
                    'isFloating': False, 'isAllDay': True, 'reminders': [], 'exDate': [],
                    'completedTime': '2018-08-09T07:20:11.000+0000', 'completedUserId': 185769383,
                    'priority': 0, 'status': 2, 'items': [], 'progress': 0,
                    'modifiedTime': '2021-01-14T08:21:01.000+0000', 'etag': 'mhjyig4y',
                    'deleted': 0, 'createdTime': '2021-01-14T08:20:10.000+0000', 'creator': 185769383, 'kind': 'TEXT'},

                    {'id': '5ffffeaab04b355792c79d89', 'projectId': 'inbox115781412',
                    'sortOrder': -6597069766656, 'title': 'Read Book', 'content': '',
                    'startDate': '2021-01-14T08:00:00.000+0000', 'dueDate': '2021-01-14T08:00:00.000+0000',
                    'timeZone': 'America/Los_Angeles', 'isFloating': False, 'isAllDay': True, 'reminders': [],
                    'exDate': [], 'completedTime': '2018-08-08T07:20:12.000+0000', 'completedUserId': 185769383,
                    'priority': 0, 'status': 2, 'items': [], 'progress': 0,
                    'modifiedTime': '2021-01-14T08:20:46.000+0000', 'etag': 'tzd4coms', 'deleted': 0,
                    'createdTime': '2021-01-14T08:19:54.000+0000', 'creator': 185769383, 'kind': 'TEXT'}]
                    ```

                    ![image](https://user-images.githubusercontent.com/56806733/104563478-8c5ea880-55fe-11eb-9bcf-91bc44c02083.png)

            === "Completed Tasks Over A Specific Duration Of Time"
                You can also get completed tasks that were completed in a specific time duration.
                Include specific hours, minutes, and seconds for the datetime objects, and
                specify `full` to be false -> meaning that the specific times will be put into effect.

                ```python
                # Get the tasks completed between 12PM and 5PM on 12/15/2020
                start = datetime(2020, 12, 15, 12)  # 12PM 12/15/2020
                end = datetime(2020, 12, 15, 17)    # 5PM 12/15/2020
                tasks = client.task.get_completed(start, end, full=False)
                ```
        """
        url = self._client.BASE_URL + "project/all/completed"

        if tz is None:
            tz = self._client.time_zone

        if not isinstance(start, datetime.datetime):
            raise TypeError("Start Must Be A Datetime Object")

        if not isinstance(end, datetime.datetime) and end is not None:
            raise TypeError("End Must Be A Datetime Object")

        # Handles case when start_date occurs after end_date
        if end is not None and start > end:
            raise ValueError("Invalid Date Range: Start Date Occurs After End Date")

        # Handles invalid timezone argument
        if tz not in pytz.all_timezones_set:
            raise KeyError("Invalid Time Zone")

        # Single Day Entry
        if end is None:
            start = datetime.datetime(start.year, start.month, start.day, 0, 0, 0)
            end = datetime.datetime(start.year, start.month, start.day, 23, 59, 59)

        # Multi DAy -> Full Day Entry
        elif full is True and end is not None:
            start = datetime.datetime(start.year, start.month, start.day, 0, 0, 0)
            end = datetime.datetime(end.year, end.month, end.day, 23, 59, 59)

        # Convert Local Time to UTC time based off the time_zone string specified
        start = convert_local_time_to_utc(start, tz)
        end = convert_local_time_to_utc(end, tz)

        parameters = {
            "from": start.strftime(DATE_FORMAT),
            "to": end.strftime(DATE_FORMAT),
            "limit": 100,
        }
        response = self._client.http_get(
            url, params=parameters, cookies=self._client.cookies, headers=self.headers
        )
        return response

    def dates(self, start, due=None, tz=None):
        """
        Performs necessary date conversions from datetime objects to strings. This
        method allows for more natural input of data to the [`builder`][managers.tasks.TaskManager.builder]
        method.

        Arguments:
            start (datetime): Desired start time
            due (datetime): Desired end time
            tz (str): Time zone string if the desired time zone is not the account default.

        Returns:
            dict: Contains 'startDate', 'endDate', 'timeZone', and 'allDay' when applicable.


        1. All Day Start Time (single day task)
        2. All Day Start and End Time (multi-day range)
        3. Specific Start Time (specific time task)
        4. Specific Start and End Time (specific start and end task)

        !!! example "Last Day Of The Month"
            ```python
            start = datetime(2027, 3, 27)
            end = datetime(2027, 3, 31)

            dates = client.task.dates(start, end)
            ```

            ??? success "Result"
                ```
                {'startDate': '2027-03-27T07:00:00+0000',
                'dueDate': '2027-04-01T07:00:00+0000',
                'allDay': True}
                ```
        """
        dates = {}
        # Set time zone
        if tz is not None:
            dates["timeZone"] = tz
        else:
            tz = self._client.time_zone

        # Check if just start date
        if due is None:
            if (
                start.hour != 0
                or start.minute != 0
                or start.second != 0
                or start.microsecond != 0
            ):
                dates["startDate"] = convert_date_to_tick_tick_format(start, tz)
                dates["allDay"] = False
            else:
                dates["startDate"] = convert_date_to_tick_tick_format(start, tz)
                dates["allDay"] = True
            return dates

        # Check all day for both
        if (
            start.hour != 0
            or start.minute != 0
            or start.second != 0
            or start.microsecond != 0
            or due.hour != 0
            or due.minute != 0
            or due.second != 0
            or due.microsecond != 0
        ):
            # Just convert the dates and return
            dates["startDate"] = convert_date_to_tick_tick_format(start, tz)
            dates["dueDate"] = convert_date_to_tick_tick_format(due, tz)
            dates["allDay"] = False
            return dates

        # All day is true, however normally right now if we were to use a date like Jan 1 - Jan 3,
        # TickTick would create a task that is only Jan 1 - Jan 2 since the date would be up to Jan 3
        # Lets account for that by making the date actually be one more than the current end date
        # This will allow for more natural date input for all day tasks
        days = monthrange(due.year, due.month)
        if due.day + 1 > days[1]:  # Last day of the month
            if due.month + 1 > 12:  # Last month of the year
                year = due.year + 1  # Both last day of month and last day of year
                day = 1
                month = 1
            else:  # Not last month of year, just reset the day and increment the month
                year = due.year
                month = due.month + 1
                day = 1
        else:  # Dont have to worry about incrementing year or month
            year = due.year
            day = due.day + 1
            month = due.month

        due = datetime.datetime(year, month, day)  # No hours, mins, or seconds needed

        dates["startDate"] = convert_date_to_tick_tick_format(start, tz)
        dates["dueDate"] = convert_date_to_tick_tick_format(due, tz)
        dates["allDay"] = True

        return dates

    def builder(
        self,
        title: str = "",
        projectId: str = None,
        content: str = None,
        desc: str = None,
        allDay: bool = None,
        startDate: datetime.datetime = None,
        dueDate: datetime.datetime = None,
        timeZone: str = None,
        reminders: list = None,
        repeat: str = None,
        priority: int = None,
        sortOrder: int = None,
        items: list = None,
    ):
        """
        Builds a task dictionary with the passed fields. This is a helper
        method for task creation.

        Arguments:
            title (str): Desired name of the task
            projectId (str): ID string of the project
            content (str): Content body of the task
            desc (str): Description of the task checklist
            allDay (bool): Boolean for whether the task is all day or not
            startDate (datetime.datetime): Start time of the task
            dueDate (datetime.datetime): End time of the task
            timeZone (str): Time zone for the task
            reminders (list): List of reminder triggers
            repeat (str): Recurring rules for the task
            priority (int): None:0, Low:1, Medium:3, High5
            sortOrder (int): Task sort order
            items (list): Subtasks of task

        Returns:
            dict: A dictionary containing the fields necessary for task creation.

        !!! example
            Building a local task object with a title, start, and due time.

            ```python
            start = datetime(2027, 5, 2)
            end = datetime(2027, 5, 7)
            title = 'Festival'
            task_dict = client.task.builder(title, startDate=start, dueDate=end)
            ```

            ??? Result

                ```python
                {'startDate': '2027-05-02T07:00:00+0000',
                 'dueDate': '2027-05-08T07:00:00+0000',
                 'allDay': True,
                 'title': 'Festival'}
                ```
        """
        task = {"title": title}
        if projectId is not None:
            task["projectId"] = projectId
        if content is not None:
            task["content"] = content
        if desc is not None:
            task["desc"] = desc
        if allDay is not None:
            task["allDay"] = allDay
        if reminders is not None:
            task["reminders"] = reminders
        if repeat is not None:
            task["repeat"] = repeat
        if priority is not None:
            task["priority"] = priority
        if sortOrder is not None:
            task["sortOrder"] = sortOrder
        if items is not None:
            task["items"] = items

        dates = {}

        # date conversions
        if startDate is not None:
            dates = self.dates(startDate, dueDate, timeZone)

        # merge dicts
        return {**dates, **task}



import argparse
from pathlib import Path

ffmpeg_parser = argparse.ArgumentParser(
    # prog=sys.argv[0],
    prog="py -m rob.ffmpeg",
    allow_abbrev=True,
    add_help=True,
    description="Concatenates audio files to a single .m4b using ffmpeg.",
    epilog="(C) Rob",
)

# Add the arguments
ffmpeg_parser.add_argument(
    "-f",
    "--filetype",
    metavar="filetype",
    nargs="+",
    default=[".mp3", ".m4a", ".ogg"],
    action="store",
    type=str,
    help="the filetype to concatenate (valid options: '.mp3', '.ogg' or '.wav')",
)

ffmpeg_parser.add_argument(
    "-p",
    "--path",
    metavar="path",
    nargs=1,
    default=".",
    action="store",
    type=str,
    help="the path to files to be concatted",
)

ffmpeg_parser.add_argument(
    "-c",
    "--cpus",
    metavar="cpus",
    default=1,
    action="store",
    type=int,
    help="the number of processor cores to utilize",
)

ffmpeg_parser.add_argument(
    "-n",
    "--number",
    metavar="number",
    default=0,
    action="store",
    type=int,
    help="the number of folders to limit execution on",
)

ffmpeg_parser.add_argument(
    "-s",
    "--start",
    metavar="start_number",
    default=0,
    action="store",
    type=int,
    help="start execution on Nth folder (ignoring previous)",
)

ffmpeg_parser.add_argument(
    "--safe",
    action="store_true",
    default=False,
    help="force ffmpeg to ask for overwrite permissions",
)

ffmpeg_parser.add_argument(
    "--command",
    action="store_true",
    help="Only print the corresponding ffmpeg command(s)",
)

ffmpeg_parser.add_argument(
    "-i", "--interact", action="store_true", help="supply arguments manually via prompt"
)



import glob
import sys
from pathlib import Path

import rich
import toml

_EXCLUSIONS = ["__init__", "__main__"]


def main():

    tom = toml.load(Path(__file__).parent / "pyproject.toml")
    version = tom["tool"]["poetry"]["version"]
    package_name = tom["tool"]["poetry"]["name"]

    rich.print(
        f"[red]Rob's Utility Scripts[/red]\n[green]Version: {version}[/green]\n\nAvailable scripts (pass -h for CLI usage):"
    )

    scripts = []
    for elem in glob.glob(f"{Path(__file__).parent}/*.py"):
        if Path(elem).stem not in _EXCLUSIONS:
            scripts.append("rob." + Path(elem).stem)

    print("\n".join(elem for elem in scripts))
    sys.stdout = None
    sys.stderr = None
    # print(
    #     "\n".join(
    #         [
    #             "rob." + Path(elem).stem
    #             for elem in glob.glob(f"{Path(__file__).parent}/*.py")
    #             if Path(elem).stem not in _EXCLUSIONS
    #         ]
    #     )
    # )


if __name__ == "__main__":
    main()



from .managers.focus import FocusTimeManager
from .managers.habits import HabitManager
from .managers.pomo import PomoManager
from .managers.projects import ProjectManager
from .managers.settings import SettingsManager
from .managers.tags import TagsManager
from .managers.tasks import TaskManager
from .oauth2 import OAuth2


class TickTickClient:
    BASE_URL = "https://api.ticktick.com/api/v2/"

    OPEN_API_BASE_URL = "https://api.ticktick.com"

    INITIAL_BATCH_URL = BASE_URL + "batch/check/0"

    USER_AGENT = "Mozilla/5.0 (Macintosh; Intel Mac OS X 10.15; rv:95.0) Gecko/20100101 Firefox/95.0"

    HEADERS = {"User-Agent": USER_AGENT}

    def __init__(self, username: str, password: str, oauth: OAuth2) -> None:
        """
        Initializes a client session. In order to interact with the API
        a successful login must occur.

        Arguments:
            username: TickTick Username
            password: TickTick Password
            oauth: OAuth2 manager

        Raises:
            RunTimeError: If the login was not successful.
        """
        # Class members

        self.access_token = None
        self.cookies = {}
        self.time_zone = ""
        self.profile_id = ""
        self.inbox_id = ""
        self.state = {}
        self.reset_local_state()
        self.oauth_manager = oauth
        self._session = self.oauth_manager.session

        self._prepare_session(username, password)

        # Mangers for the different operations
        self.focus = FocusTimeManager(self)
        self.habit = HabitManager(self)
        self.project = ProjectManager(self)
        self.pomo = PomoManager(self)
        self.settings = SettingsManager(self)
        self.tag = TagsManager(self)
        self.task = TaskManager(self)

    def _prepare_session(self, username, password):
        """
        Creates all the necessary calls to prepare the session
        """
        self._login(username, password)
        self._settings()
        self.sync()

    def reset_local_state(self):
        """
        Resets the contents of the items in the [`state`](api.md#state) dictionary.

        """
        self.state = {
            "projects": [],
            "project_folders": [],
            "tags": [],
            "tasks": [],
            "user_settings": {},
            "profile": {},
        }

    def _login(self, username: str, password: str) -> None:
        """
        Logs in to TickTick and sets the instance access token.

        Arguments:
            username: TickTick Username
            password: TickTick Password

        """
        url = self.BASE_URL + "user/signin"
        user_info = {"username": username, "password": password}
        parameters = {"wc": True, "remember": True}

        response = self.http_post(
            url, json=user_info, params=parameters, headers=self.HEADERS
        )

        self.access_token = response["token"]
        self.cookies["t"] = self.access_token

    @staticmethod
    def check_status_code(response, error_message: str) -> None:
        """
        Verifies the http response was status code 200.

        Arguments:
            response (httpx): Httpx response
            error_message: Error message to be included with the exception

        Raises:
            RuntimeError: If the status code of the response was not 200.
        """
        if response.status_code != 200:
            raise RuntimeError(error_message)

    def _settings(self):
        """
        Sets the time_zone and profile_id.

        Returns:
            The httpx response object.
        :return: httpx object containing the response from the get request
        """
        url = self.BASE_URL + "user/preferences/settings"
        parameters = {"includeWeb": True}
        response = self.http_get(
            url, params=parameters, cookies=self.cookies, headers=self.HEADERS
        )

        self.time_zone = response["timeZone"]
        self.profile_id = response["id"]

        return response

    def sync(self):
        """
        Populates the `TickTickClient` [`state`](api.md#state) dictionary with the contents of your account.

        **This method is called when necessary by other methods and does not need to be explicitly called.**

        Returns:
            httpx: The response from the get request.

        Raises:
            RunTimeError: If the request could not be completed.
        """
        response = self.http_get(
            self.INITIAL_BATCH_URL, cookies=self.cookies, headers=self.HEADERS
        )

        # Inbox Id
        self.inbox_id = response["inboxId"]
        # Set list groups
        self.state["project_folders"] = response["projectGroups"]
        # Set lists
        self.state["projects"] = response["projectProfiles"]
        # Set Uncompleted Tasks
        self.state["tasks"] = response["syncTaskBean"]["update"]
        # Set tags
        self.state["tags"] = response["tags"]

        return response

    def http_post(self, url, **kwargs):
        """
        Sends an http post request with the specified url and keyword arguments.

        Arguments:
            url (str): Url to send the request.
            **kwargs: Arguments to send with the request.

        Returns:
            dict: The json parsed response if possible or just a string of the response text if not.

        Raises:
            RunTimeError: If the request could not be completed.
        """
        response = self._session.post(url, **kwargs)
        self.check_status_code(response, "Could Not Complete Request")

        try:
            return response.json()
        except ValueError:
            return response.text

    def http_get(self, url, **kwargs):
        """
        Sends an http get request with the specified url and keyword arguments.

        Arguments:
            url (str): Url to send the request.
            **kwargs: Arguments to send with the request.

        Returns:
            dict: The json parsed response if possible or just a string of the response text if not.

        Raises:
            RunTimeError: If the request could not be completed.
        """
        response = self._session.get(url, **kwargs)
        self.check_status_code(response, "Could Not Complete Request")

        try:
            return response.json()
        except ValueError:
            return response.text

    def http_delete(self, url, **kwargs):
        """
        Sends an http delete request with the specified url and keyword arguments.

        Arguments:
            url (str): Url to send the request.
            **kwargs: Arguments to send with the request.

        Returns:
            dict: The json parsed response if possible or just a string of the response text if not.

        Raises:
            RunTimeError: If the request could not be completed.
        """
        response = self._session.delete(url, **kwargs)
        self.check_status_code(response, "Could Not Complete Request")

        try:
            return response.json()
        except ValueError:
            return response.text

    def http_put(self, url, **kwargs):
        """
        Sends an http put request with the specified url and keyword arguments.

        Arguments:
            url (str): Url to send the request.
            **kwargs: Arguments to send with the request.

        Returns:
            dict: The json parsed response if possible or just a string of the response text if not.

        Raises:
            RunTimeError: If the request could not be completed.
        """
        response = self._session.put(url, **kwargs)
        self.check_status_code(response, "Could Not Complete Request")

        try:
            return response.json()
        except ValueError:
            return response.text

    @staticmethod
    def parse_id(response: dict) -> str:
        """
        Parses the Id of a successful creation of a TickTick object.
        !!! info
            The response from the TickTick servers is in this form:

            ```md
            {'id2etag': {'5ff2bcf68f08093e5b745a30': '3okkc2xm'}, 'id2error': {}}
            ```
            We want to obtain '5ff2bcf68f08093e5b745a30' in this example - the id of the object.

        Arguments:
            response: Dictionary containing the Dd from the TickTick servers.

        Returns:
            Id string of the object.
        """
        id_tag = response["id2etag"]
        id_tag = list(id_tag.keys())
        return id_tag[0]

    @staticmethod
    def parse_etag(response: dict, multiple: bool = False) -> str:
        """
        Parses the etag of a successful creation of a tag object.

        !!! info
            The response from TickTick upon a successful tag creation is in this form:

            ```md
            {"id2etag":{"MyTag":"vxzpwo38"},"id2error":{}}
            ```
            We want to obtain "vxzpwo38" in this example - the etag of the object.

        Arguments:
            response: Dictionary from the successful creation of a tag object
            multiple: Specifies whether there are multiple etags to return.

        Return:
            A single etag string if not multiple, or a list of etag strings if multiple.
        """
        etag = response["id2etag"]
        etag2 = list(etag.keys())
        if not multiple:
            return etag[etag2[0]]
        else:
            etags = []
            for key in range(len(etag2)):
                etags.append(etag[etag2[key]])
            return etags

    def get_by_fields(self, search: str = None, **kwargs):
        """
        Finds and returns the objects in `state` that match the inputted fields.

        If search is specified, it will only search the specific [`state`](api.md#state) list,
        else the entire [`state`](api.md#state) dictionary will be searched.

        !!! example
            Since each TickTick object like tasks, projects, and tags are just dictionaries of fields,
            we can find an object by
            comparing any fields contained in those objects.

            For example: Lets say we have 3 task objects that are titled 'Hello', and we want to obtain all of them.

            The call to the function would look like this:

            ```python
            # Assumes that `client` is the name referencing the TickTickClient instance.

            found_objs = client.get_by_fields(title='Hello')
            ```
            `found_objs` would now reference a list containing the 3 objects with titles 'Hello'.

            Furthermore if we know the type of object we are looking for, we can make the search more efficient by
            specifying the key its located under in the [`state`](#state) dictionary.

            ```python
            # Assumes that `client` is the name referencing the TickTickClient instance.

            found_obj = client.get_by_fields(title='Hello', search='tasks')
            ```

            The search will now only look through `tasks` in [`state`](api.md#state).


        Arguments:
            search: Key in [`state`](api.md#state) that the search should take place in. If empty the
            entire [`state`](api.md#state) dictionary will be searched.
            **kwargs: Matching fields in the object to look for.

        Returns:
            dict or list:
            **Single Object (dict)**: The dictionary of the object.

            **Multiple Objects (list)**: A list of dictionary objects.

            **Nothing Found (list)**: Empty List

        Raises:
            ValueError: If no key word arguments are provided.
            KeyError: If the search key provided is not a key in `state`.
        """
        if kwargs == {}:
            raise ValueError("Must Include Field(s) To Be Searched For")

        if search is not None and search not in self.state:
            raise KeyError(f"'{search}' Is Not Present In self.state Dictionary")

        objects = []
        if search is not None:
            # If a specific key was passed for self.state
            # Go through self.state[key_name] and see if all the fields in kwargs match
            # If all don't match return empty list
            for index in self.state[search]:
                all_match = True
                for field in kwargs:
                    if kwargs[field] != index[field]:
                        all_match = False
                        break
                if all_match:
                    objects.append(index)

        else:
            # No key passed, search entire self.state dictionary
            # Search the first level of the state dictionary
            for primarykey in self.state:
                skip_primary_key = False
                all_match = True
                middle_key = 0
                # Search the individual lists of the dictionary
                for middle_key in range(len(self.state[primarykey])):
                    if skip_primary_key:
                        break
                    # Match the fields in the kwargs dictionary to the specific object -> if all match add index
                    for fields in kwargs:
                        # if the field doesn't exist, we can assume every other item in the list doesn't have the
                        # field either -> so skip this primary_key entirely
                        if fields not in self.state[primarykey][middle_key]:
                            all_match = False
                            skip_primary_key = True
                            break
                        if kwargs[fields] == self.state[primarykey][middle_key][fields]:
                            all_match = True
                        else:
                            all_match = False
                    if all_match:
                        objects.append(self.state[primarykey][middle_key])

        if len(objects) == 1:
            return objects[0]
        else:
            return objects

    def get_by_id(self, obj_id: str, search: str = None) -> dict:
        """
        Returns the dictionary of the object corresponding to the passed id.

        If search is specified, it will only search the specific [`state`](api.md#state) list, else the
        entire [`state`](api.md#state) dictionary will be searched.


        !!! example
            Since each TickTick object like tasks, projects, and tags are just dictionaries of fields,
            we can find an object by
            comparing the id fields.

            For example: Lets get the object that corresponds to an id referenced by `my_id`.

            The call to the function would look like this:

            ```python
            # Assumes that `client` is the name referencing the TickTickClient instance.

            found_obj = client.get_by_id(my_id)
            ```
            `found_obj` would now reference the object if it was found, else it would reference an empty dictionary.

            Furthermore if we know the type of object we are looking for, we can make the search more efficient by
            specifying the key its located under in the [`state`](api.md#state) dictionary.

            ```python
            # Assumes that `client` is the name referencing the TickTickClient instance.

            found_obj = client.get_by_id(my_id, search='projects')
            ```

            The search will now only look through `projects` in [`state`](api.md#state).

        Arguments:
            obj_id: Id of the item.
            search: Key in [`state`](api.md#state) that the search should take place in. If empty the
            entire [`state`](api.md#state) dictionary will be searched.

        Returns:
            The dictionary object of the item if found, or an empty dictionary if not found.

        Raises:
            KeyError: If the search key provided is not a key in [`state`](api.md#state).
        """
        if search is not None and search not in self.state:
            raise KeyError(f"'{search}' Is Not Present In self.state Dictionary")

        # Search just in the desired list
        if search is not None:
            for index in self.state[search]:
                if index["id"] == obj_id:
                    return index

        else:
            # Search all items in self.state
            for prim_key in self.state:
                for our_object in self.state[prim_key]:
                    if "id" not in our_object:
                        break
                    if our_object["id"] == obj_id:
                        return our_object
        # Return empty dictionary if not found
        return {}

    def get_by_etag(self, etag: str, search: str = None) -> dict:
        """
        Returns the dictionary object of the item with the matching etag.

        If search is specified, it will only search the specific [`state`](api.md#state) list, else the
        entire [`state`](api.md#state) dictionary will be searched.

        !!! example
            Since each TickTick object like tasks, projects, and tags are just dictionaries of fields,
            we can find an object by
            comparing the etag fields.

            For example: Lets get the object that corresponds to an etag referenced by `my_etag`.

            The call to the function would look like this:

            ```python
            # Assumes that `client` is the name referencing the TickTickClient instance.

            found_obj = client.get_by_etag(my_etag)
            ```
            `found_obj` would now reference the object if it was found, else it would reference an empty dictionary.

            Furthermore if we know the type of object we are looking for, we can make the search more efficient by
            specifying the key its located under in the [`state`](api.md#state) dictionary.

            ```python
            # Assumes that `client` is the name referencing the TickTickClient instance.

            found_obj = client.get_by_etag(my_etag, search='projects')
            ```

            The search will now only look through `projects` in [`state`](api.md#state).

        Arguments:
            etag: The etag of the object that you are looking for.
            search: Key in [`state`](#state) that the search should take place in. If empty the
            entire [`state`](api.md#state) dictionary will be searched.

        Returns:
            The dictionary object of the item if found, or an empty dictionary if not found.

        Raises:
            KeyError: If the search key provided is not a key in [`state`](api.md#state).

        """
        if search is not None and search not in self.state:
            raise KeyError(f"'{search}' Is Not Present In self.state Dictionary")

        # Search just in the desired list
        if search is not None:
            for index in self.state[search]:
                if index["etag"] == etag:
                    return index

        else:
            # Search all items in self.state
            for prim_key in self.state:
                for our_object in self.state[prim_key]:
                    if "etag" not in our_object:
                        break
                    if our_object["etag"] == etag:
                        return our_object
        # Return empty dictionary if not found
        return {}

    def delete_from_local_state(self, search: str = None, **kwargs) -> dict:
        """
        Deletes a single object from the local `state` dictionary. **Does not delete any items remotely.**

        If search is specified, it will only search the specific [`state`](api.md#state) list,
        else the entire [`state`](api.md#state) dictionary will be searched.

        !!! example
            Since each TickTick object like tasks, lists, and tags are just dictionaries of fields,
            we can find an object by
            comparing the fields.

            For example: Lets say that we wanted to find and delete an existing task object from our local state
            with the name 'Get Groceries'. To do this, we can specify the field(s) that we want to compare for in
            the task objects -> in this case the `title` 'Get Groceries'.

            The call to the function would look like this:

            ```python
            # Assumes that `client` is the name referencing the TickTickClient instance.

            deleted_task = client.delete_from_local_state(title='Get Groceries')
            ```
            `deleted_task` would now hold the object that was deleted from the [`state`](api.md#state)
            dictionary if it was found.

            Furthermore if we know the type of object we are looking for, we can make the search more efficient by
            specifying the key its located under in the [`state`](api.md#state) dictionary.

            ```python
            # Assumes that `client` is the name referencing the TickTickClient instance.

            deleted_task = client.delete_from_local_state(title='Get Groceries', search='tasks')
            ```

            The search will now only look through `tasks` in `state`.


        Arguments:
            search: A specific item to look through in the [`state`](api.md#state) dictionary. When not specified the
            entire [`state`](api.md#state) dictionary will be searched.
            **kwargs: Matching fields in the object to look for.

        Returns:
            The dictionary of the object that was deleted.

        Raises:
            ValueError: If no key word arguments are provided.
            KeyError: If the search key provided is not a key in [`state`](api.md#state).

        """
        # Check that kwargs is not empty
        if kwargs == {}:
            raise ValueError("Must Include Field(s) To Be Searched For")

        if search is not None and search not in self.state:
            raise KeyError(f"'{search}' Is Not Present In self.state Dictionary")

        # Search just in the desired list
        if search is not None:
            # Go through the state dictionary list and delete the object that matches the fields
            for item in range(len(self.state[search])):
                all_match = True
                for field in kwargs:
                    if kwargs[field] != self.state[search][item][field]:
                        all_match = False
                        break
                if all_match:
                    deleted = self.state[search][item]
                    # Delete the item
                    del self.state[search][item]
                    return deleted

        else:
            # No key passed, search entire self.state dictionary
            # Search the first level of the state dictionary
            for primary_key in self.state:
                skip_primary_key = False
                all_match = True
                middle_key = 0
                # Search the individual lists of the dictionary
                for middle_key in range(len(self.state[primary_key])):
                    if skip_primary_key:
                        break
                    # Match the fields in the kwargs dictionary to the specific object -> if all match add index
                    for fields in kwargs:
                        # if the field doesn't exist, we can assume every other item in the list doesn't have the
                        # field either -> so skip this primary_key entirely
                        if fields not in self.state[primary_key][middle_key]:
                            all_match = False
                            skip_primary_key = True
                            break
                        if (
                            kwargs[fields]
                            == self.state[primary_key][middle_key][fields]
                        ):
                            all_match = True
                        else:
                            all_match = False
                    if all_match:
                        deleted = self.state[primary_key][middle_key]
                        del self.state[primary_key][middle_key]
                        return deleted



import deal
class HabitManager:
    @deal.pure
    def __init__(self, client_class):
        self._client = client_class
        self.access_token = ""

    #   ---------------------------------------------------------------------------------------------------------------
    #   Habit Methods

    @deal.pure
    def get_checkins(self):
        # POST habitChekckins/query
        """
        {
            'postData': '{"habitIds":["",""],"afterStamp":20211224}',
            'hasPostData': True,
            'postDataEntries': [
                {
                    'bytes' b64encode(postData)
                }
            ]
        }
        """
        pass

    @deal.pure
    def batch_checkins(self, data, function):
        # POST habitCheckins/batch (create)
        """
        postData:  '{
            "add":[
                {
                    "checkinStamp": 20211201,
                    "checkinTime": "2021-12-31T07:30:37.000+0000",
                    "goal":1,
                    "habitId: "safasfas0990",
                    "id": "f90sdag0asd9f0as",
                    "status":2,
                    "value":1
                }
            ],
            "update":[],
            "delete":[]
        }',
        hasPostData: true,
        postDataEntries: [
            {
                bytes: base64encode(postData)
            }
        ]
        """
        # POST habitCheckins/batch (update)
        """
        postData: '{"add":[],"update":[{"checkinStamp":20211201,"checkinTime":"2021-12-31T07:36:45.000+0000","goal":1,"habitId":"6038980b801c81216396971b","id":"61ceb195ac8b5d3c97e7a7de","status":0,"value":0}],"delete":[]}',
        hasPostData: true,
        postDataEntries: [
            {
                bytes: 'eyJhZGQiOltdLCJ1cGRhdGUiOlt7ImNoZWNraW5TdGFtcCI6MjAyMTEyMDEsImNoZWNraW5UaW1lIjoiMjAyMS0xMi0zMVQwNzozNjo0NS4wMDArMDAwMCIsImdvYWwiOjEsImhhYml0SWQiOiI2MDM4OTgwYjgwMWM4MTIxNjM5Njk3MWIiLCJpZCI6IjYxY2ViMTk1YWM4YjVkM2M5N2U3YTdkZSIsInN0YXR1cyI6MCwidmFsdWUiOjB9XSwiZGVsZXRlIjpbXX0='
            }
        ],
        """
        pass

    @deal.pure
    def add_checkins(self, data):
        self.batch_checkins(data, "add")
        pass

    @deal.pure
    def update_checkins(self, data):
        self.batch_checkins(data, "update")
        pass

    @deal.pure
    def delete_checkins(self, data):
        self.batch_checkins(data, "delete")
        pass

    @deal.pure
    def get_habits(self):
        # GET habits
        pass

    @deal.pure
    def batch_habits(self, data, function):
        # POST habits/batch (update)
        """
        postData: '{"add":[],"update":[{"color":"#F8D550","iconRes":"habit_early_to_rise","createdTime":"2021-02-26T06:41:15.000+0000","encouragement":"Get up and be amazing","etag":"2peywvxk","goal":1,"id":"6038980b801c81216396971b","modifiedTime":"2021-06-01T15:06:35.000+0000","name":"Early to Rise","recordEnable":false,"reminders":["06:30"],"repeatRule":"RRULE:FREQ=WEEKLY;INTERVAL=1;BYDAY=SU,MO,TU,WE,TH,FR,SA","sortOrder":-4611686430744248000,"status":0,"step":0,"totalCheckIns":2,"type":"Boolean","unit":"Count","sectionId":"-1","targetDays":0,"targetStartDate":0,"completedCycles":0}],"delete":[]}',
        hasPostData: true,
        postDataEntries: [
            {
            bytes: 'eyJhZGQiOltdLCJ1cGRhdGUiOlt7ImNvbG9yIjoiI0Y4RDU1MCIsImljb25SZXMiOiJoYWJpdF9lYXJseV90b19yaXNlIiwiY3JlYXRlZFRpbWUiOiIyMDIxLTAyLTI2VDA2OjQxOjE1LjAwMCswMDAwIiwiZW5jb3VyYWdlbWVudCI6IkdldCB1cCBhbmQgYmUgYW1hemluZyIsImV0YWciOiIycGV5d3Z4ayIsImdvYWwiOjEsImlkIjoiNjAzODk4MGI4MDFjODEyMTYzOTY5NzFiIiwibW9kaWZpZWRUaW1lIjoiMjAyMS0wNi0wMVQxNTowNjozNS4wMDArMDAwMCIsIm5hbWUiOiJFYXJseSB0byBSaXNlIiwicmVjb3JkRW5hYmxlIjpmYWxzZSwicmVtaW5kZXJzIjpbIjA2OjMwIl0sInJlcGVhdFJ1bGUiOiJSUlVMRTpGUkVRPVdFRUtMWTtJTlRFUlZBTD0xO0JZREFZPVNVLE1PLFRVLFdFLFRILEZSLFNBIiwic29ydE9yZGVyIjotNDYxMTY4NjQzMDc0NDI0ODAwMCwic3RhdHVzIjowLCJzdGVwIjowLCJ0b3RhbENoZWNrSW5zIjoyLCJ0eXBlIjoiQm9vbGVhbiIsInVuaXQiOiJDb3VudCIsInNlY3Rpb25JZCI6Ii0xIiwidGFyZ2V0RGF5cyI6MCwidGFyZ2V0U3RhcnREYXRlIjowLCJjb21wbGV0ZWRDeWNsZXMiOjB9XSwiZGVsZXRlIjpbXX0='
            }
        """
        pass

    @deal.pure
    def update_habits(self, data):
        self.batch_habits(data, "update")
        pass

    @deal.pure
    def add_habits(self, data):
        self.batch_habits(data, "add")
        pass

    @deal.pure
    def delete_habits(self, data):
        self.batch_habits(data, "delete")
        pass

    @deal.pure
    def get_sections(self):
        # GET habitSections
        pass

    @deal.pure
    def get_specific_records(self, ids=[], afterstamp=""):
        # GET habitRecords?habitIds=6037eb14801c327339afbd3a&habitIds=603897ae0cda91d09c9f376f&habitIds=6038980b801c81216396971b&habitIds=60389839801c81216396971c&afterStamp=20211230
        pass


# Other endpoints
# OPTIONS habitCheckins/query



import argparse

reddit_parser = argparse.ArgumentParser(
    prog="py -m rob.reddit_archive",
    allow_abbrev=True,
    add_help=True,
    description="Archive reddit comments.",
    epilog="(C) Rob",
)

reddit_parser.add_argument(
    "-t", "--text", action="store_true", help="generate text file ONLY"
)
reddit_parser.add_argument(
    "-o", "--overwrite", action="store_true", help="overwrite existing database entries"
)
reddit_parser.add_argument(
    "-u",
    "--user",
    metavar="user",
    nargs=1,
    action="store",
    type=str,
    help="specify user",
)

reddit_parser.add_argument(
    "-p",
    "--password",
    metavar="password",
    nargs=1,
    action="store",
    type=str,
    help="specify password",
)

reddit_parser.add_argument(
    "-f",
    "--full",
    action="store_true",
    help="Include 'top' and 'controversial' posts in praw request.",
)

reddit_parser.add_argument(
    "-c",
    "--config",
    action="store_true",
    help="Manually configure reddit credentials for this user agent.",
)

reddit_parser.add_argument(
    "--csv",
    metavar="csv_file",
    nargs=1,
    action="store",
    type=str,
    help="specify csv file to parse and archive",
)

reddit_parser.add_argument(
    "--interact",
    action="store_true",
    help="interact with database and reddit API",
)

reddit_parser.add_argument(
    "--no-text",
    action="store_true",
    help="Suppress text file generation.",
)



import glob
import multiprocessing
import os
import shutil
import subprocess
import sys
from pathlib import Path
from typing import List

import rich
import rich.traceback
import taglib
from pydub.utils import mediainfo
from rich import pretty
from rich.progress import track

from .parser.ffmpeg_parser import ffmpeg_parser

pretty.install()
rich.traceback.install()

_ARGS = ffmpeg_parser.parse_args()

_PATH = Path(_ARGS.path)
_FILETYPES = _ARGS.filetype
_SAFE = _ARGS.safe
_CPUS = _ARGS.cpus
_COMMAND = _ARGS.command

_PROMPT = f"rob.{Path(__file__).stem}> "
_PROMPT_STYLE = "white on blue"
_ERROR_STYLE = "red on black"
_COMMAND_FILE = Path(os.getcwd()) / "ffmpeg_commands.ps1"


def command_only(folder: Path) -> None:
    """Generate a textfile containing the desired ffmpeg CLI commands."""
    os.chdir(folder)

    mp3s = []
    for filetype in _FILETYPES:
        temp = glob.glob(f"*{filetype}")
        mp3s.extend([Path(elem) for elem in temp if elem[0] != "~"])

    if not mp3s:
        return

    for i, mp3 in enumerate(mp3s):
        if "'" in str(mp3.stem):
            shutil.move(mp3, str(mp3).replace("'", ""))
            mp3s[i] = Path(str(mp3).replace("'", ""))

    concated = "|".join([str(elem.absolute()) for elem in mp3s])
    command = [
        "ffmpeg",
        "-i",
        f'''"concat:{concated}"''',
        # "-movflags", # Carry metadata over
        # "use_metadata_tags",
        "-map_metadata",
        "0",  # Use metadata from 0th input
        "-c:a",  # audio cocec
        "aac",
        "-b:a",  # bitrate
        "64k",
        "-c:v",
        "copy",
        f"'{Path(mp3s[0]).stem}.m4b'",
        "-y",
    ]

    if _SAFE:
        command.remove("-y")

    with open(_COMMAND_FILE, "a") as fp:
        fp.write(" ".join(command) + "\n")

    print(" ".join(command))

    return


def concat_and_convert(folder: Path) -> None:
    """Spin up an ffmpeg process in target folder."""

    os.chdir(folder)

    global _SAFE

    mp3s = []
    for filetype in _FILETYPES:
        temp = glob.glob(f"*{filetype}")
        mp3s.extend([Path(elem) for elem in temp if elem[0] != "~"])

    if not mp3s:
        return

    for i, mp3 in enumerate(mp3s):
        if "'" in str(mp3.stem):
            shutil.move(mp3, str(mp3).replace("'", ""))
            mp3s[i] = Path(str(mp3).replace("'", ""))

    fp = open("files.txt", "w")
    for file in track(mp3s):
        fp.write(f"file '{file}'\n")

    fp.close()

    bit_rate = mediainfo(mp3s[0])["bit_rate"]

    command = [
        "ffmpeg",
        "-f",
        "concat",
        "-safe",
        "0",
        "-i",
        "files.txt",
        "-map_metadata",
        "0",  # Use metadata from 0th input
        "-movflags",  # Carry metadata over
        "use_metadata_tags",
        "-c:a",  # audio cocec
        "aac",  # audio codec
        "-b:a",  # bitrate
        f"{bit_rate}",
        "-vn",  # no video stream
        "-y",
        f"""{Path(mp3s[0]).stem}.m4b""",
    ]

    if _SAFE:
        command.remove("-y")

    rich.print("[green]Executing:\n[yellow]" + " ".join(command))
    subprocess.run(command, shell=True)

    os.remove("files.txt")

    # <Transfer metadata>
    inf = taglib.File(f"{mp3s[0]}")
    outf = taglib.File(f"{Path(mp3s[0]).stem}.m4b")

    outf.tags = inf.tags
    outf.save()
    outf.close()
    inf.close()
    # </Transfer metadata>

    return


def interact() -> List[Path]:
    """Walk the user through configuring an ffmpeg call.

    Returns:
        folders (List[Path]): the list of folders to traverse and modify.
    """
    global _PATH
    global _FILETYPES
    global _SAFE
    global _CPUS
    global _COMMAND

    final_choice = False
    valid = True
    while not final_choice:
        style = _PROMPT_STYLE if valid else _ERROR_STYLE
        rich.print(
            f"""[{style}]Enter the path to work on.[/{style}]
\[default: {_PATH.parent.absolute()}]"""
        )
        choice = input(f"{_PROMPT}")

        if Path(choice).exists():
            _PATH = Path(choice)
            final_choice = True

        valid = False

    rich.print(f"[yellow]Working on '{_PATH.absolute()}'.\n")

    rich.print(
        f"[{_PROMPT_STYLE}]Which file type to work on?[/{_PROMPT_STYLE}] \[default: {_FILETYPES}]\n 1.) *.mp3\n 2.) *.wav\n 3.) *.ogg"
    )
    choice = input(f"{_PROMPT}")

    match choice:
        case "1" | "*.mp3" | ".mp3" | "mp3":
            _FILETYPES = [".mp3"]

        case "2" | "*.wav" | ".wav" | "wav":
            _FILETYPES = [".wav"]

        case "3" | "*.ogg" | ".ogg" | "ogg":
            _FILETYPES = [".ogg"]

        case _ as x:
            if x:
                _FILETYPES = [choice]

    rich.print(f"[yellow]Working on {_FILETYPES}'s.\n")

    folders = set()
    os.chdir(_PATH)
    for filetype in _FILETYPES:
        for file in glob.glob(f"**/*{filetype}", recursive=True):
            if _SAFE and ".m4b" not in _FILETYPES:
                try:
                    os.remove(Path(f"{Path(file).parent}/{Path(file).stem}.m4b"))
                except FileNotFoundError:
                    pass
            target = Path(file).absolute()
            folders.add(target.parent)

    all_folders = sorted(list(folders))
    folders = all_folders[:]

    final_choice = False
    invalid = False
    while not final_choice:
        rich.print(f"Folders with {_FILETYPES}'s:")
        for count, folder in enumerate(all_folders):
            if folder in folders:
                rich.print(f" [green]{count+1}.) {folder}")
            else:
                rich.print(f" [red]{count+1}.) {folder}")

        style = _PROMPT_STYLE if not invalid else _ERROR_STYLE
        rich.print(
            f"\n[{style}]Toggle execution of folders by number or press Enter to continue."
        )
        choice = input(f"{_PROMPT}")
        invalid = False
        try:
            if not choice:
                final_choice = True

            choice = int(choice)
            if 1 <= choice <= len(all_folders):
                if all_folders[choice - 1] in folders:
                    folders.remove(all_folders[choice - 1])
                else:
                    folders.insert(choice - 1, all_folders[choice - 1])
            else:
                invalid = True

        except (SyntaxError, ValueError):
            invalid = True

    rich.print(
        f"[{_PROMPT_STYLE}]How many CPU cores to use?[/{_PROMPT_STYLE}] \[default: {_CPUS}]"
    )
    choice = input(f"{_PROMPT}")

    while True:
        try:
            if not choice:
                break
            else:
                _CPUS = int(choice)
                break
        except (TypeError, ValueError):
            rich.print(
                f"[red]Must supply an integer.[/red]\nHow many CPU cores to use \[default: {_CPUS}]?\n"
            )
            choice = input(f"{_PROMPT}")

    rich.print(f"[yellow]Using {_CPUS} core{'s' if _CPUS>1 else ''}.\n")

    rich.print(
        f"[{_PROMPT_STYLE}](1) Execute commands or \n(2) Generate command file?[/{_PROMPT_STYLE}]\n\[default {'1' if _COMMAND else '1'}]"
    )
    choice = input(f"{_PROMPT}")

    match choice:
        case "1" | "execute":
            _COMMAND = False
        case "2" | "generate" | "file":
            _COMMAND = True
            open(_COMMAND_FILE, "w").close()
        case _:
            pass

    return folders


def main() -> None:
    folders = set()

    if _COMMAND:
        open(_COMMAND_FILE, "w").close()

    if _ARGS.interact:
        folders = interact()

    else:
        os.chdir(_PATH)
        for filetype in _FILETYPES:
            for file in glob.glob(f"**/*{filetype}", recursive=True):
                if not _SAFE:
                    try:
                        os.remove(Path(f"{Path(file).parent}/{Path(file).stem}.m4b"))
                    except FileNotFoundError:
                        pass
                target = Path(file).absolute()
                folders.add(target.parent)

        folders = sorted(list(folders))

        # optionally slice the list of folders
        match (bool(_ARGS.number), bool(_ARGS.start)):
            case (True, False):
                folders = folders[0 : _ARGS.number]
            case (False, True):
                folders = folders[_ARGS.start - 1 :]
            case (True, True):
                folders = folders[_ARGS.start - 1 : _ARGS.start - 1 + _ARGS.number]
            case _:
                pass

    if not folders:
        rich.print(f"[{_ERROR_STYLE}]No {_FILETYPES}'s found.")
        sys.exit(1)

    rich.print(
        f"[{_PROMPT_STYLE}]{'Compiling' if _COMMAND else 'Executing'} on[/{_PROMPT_STYLE}]\n * "
        + "\n * ".join([str(folder) for folder in folders])
        + f". \n[{_PROMPT_STYLE}]Using {_CPUS} core. Ok?"
    )
    input(f"{_PROMPT}")

    if _COMMAND:
        with multiprocessing.Pool(_CPUS) as pool:
            pool.map_async(command_only, folders).get()
    else:
        with multiprocessing.Pool(_CPUS) as pool:
            pool.map_async(concat_and_convert, folders).get()


if __name__ == "__main__":
    main()



"""
Provides some methods for dealing with hex color code strings.
"""

import random
import re
import deal

VALID_HEX_VALUES = "^#([A-Fa-f0-9]{6}|[A-Fa-f0-9]{3})$"


@deal.has('random')
@deal.safe
def generate_hex_color() -> str:
    """
    Generates a random hexadecimal color string to be used for rgb color schemes.

    Returns:
        '#' followed by 6 hexadecimal digits.

    ??? info "Import Help"
        ```python
        from ticktick.helpers.hex_color import generate_hex_color
        ```
    """
    num = random.randint(1118481, 16777215)
    hex_num = format(num, 'x')
    return '#' + hex_num


@deal.pure
def check_hex_color(color: str) -> bool:
    """
    Verifies if the passed in color string is a valid hexadecimal color string

    Arguments:
        color: String to check.

    Returns:
        True if the string is a valid hex code, else False.

    ??? info "Import Help"
        ```python
        from ticktick.helpers.hex_color import check_hex_color
        ```
    """
    check_color = re.search(VALID_HEX_VALUES, color)
    if not check_color:
        return False
    else:
        return True



import deal
class PomoManager:

    @deal.pure
    def __init__(self, client_class):
        self._client = client_class
        self.access_token = ''

    @deal.pure
    def start(self):
        pass

    @deal.pure
    def statistics(self):
        # https://api.ticktick.com/api/v2/statistics/general
        pass



import argparse

clean_parser = argparse.ArgumentParser(
    description="Clean up a folder.",
    add_help=True,
    epilog="(C) Rob",
)

# Add the arguments
clean_parser.add_argument(
    "-p",
    "--path",
    metavar="dir",
    default=".",
    action="store",
    type=str,
    help="the directory to clean",
)

clean_parser.add_argument(
    "-y",
    "--yes",
    default=False,
    action="store_true",
    help="skip all interactive prompts by answering 'yes'",
)

clean_parser.add_argument(
    "-r",
    "--recurse",
    default=False,
    action="store_true",
    help="Recurse through nested folders.",
)

alt_mode = clean_parser.add_mutually_exclusive_group(required=False)

alt_mode.add_argument(
    "-u",
    "--undo",
    default=False,
    action="store_true",
    help="undo the previous execution in target directory",
)

alt_mode.add_argument(
    "--config",
    default=False,
    action="store_true",
    help="create a config file",
)

alt_mode.add_argument(
    "--debug", default=False, action="store_true", help="run in debug mode"
)
